/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "log.h"
#include "config.h"

// ***********************************************************************
// 
// ***********************************************************************
void config_defaults(config *c)
{
	c->host = "0.0.0.0";
	c->port = "9080";
	c->adev = "hw:0";
	c->arate = 48000;
	c->ablock = 0.020;
	c->athreshold = 0.0229;
	c->vdev = "/dev/video0";
	c->vwidth = 0;
	c->vheight = 0;
	c->vsleep = 20;
	c->vwakeup = 0;
	c->vmask = NULL;
	c->iwidth = 0;
	c->iheight = 0;
	c->ipixels = 0;
	c->iquality = 75;
	c->rspan = 3600 * 2;
	c->rithr = 0;
	c->rpthr = 100;
	c->rbginertia = 0.8;
	c->rfginertia = 0.5;
	c->rsettle = 20;
	c->ravgmin = 0;
	c->ravgmax = 255;
	c->rwakeup = 0;
	c->rdraw = 0;
	c->detach = 0;
	c->logfile = "/tmp/l_mon.log";
	c->sinterval = 5;
	c->pinterval = 1.0;
}


// ***********************************************************************
// 
// ***********************************************************************
#define HELP_MESSAGE	\
"Usage: l_mon [options]\n" \
"\n" \
"Network options:\n" \
"  -b host        server bind address, default %s\n" \
"  -p port        server port (service), default %s\n" \
"\n" \
"Video options:\n" \
"  -vd device     v4l2 video device, default %s\n" \
"  -vw width      device horizontal resolution, default %d px\n" \
"  -vh height     device vertical resolution, default %d px\n" \
"  -vs seconds    video device inactivity timeout, default %d s\n" \
"  -vu seconds    video device wakeup interval, default %d s\n" \
"  -vm mask       mask for input image, jpeg file, default %s\n" \
"  -iw width      image horizontal resolution, default %d px\n" \
"  -ih height     image vertical resolution, default %d px\n" \
"  -ip pixels     image size, default %d pixels\n" \
"  -iq quality    image quality, default %d %%\n" \
"\n" \
"Activity roll options:\n" \
"  -rs seconds    history span, default %d s\n" \
"  -ri value      pixel intensity threshold, default %d (/ 256)\n" \
"  -rp count      pixel count threshold, default %d (/ 65536)\n" \
"  -r0 inertia    background smoothing inertia, default %g\n" \
"  -r1 inertia    foreground smoothing inertia, default %g\n" \
"  -rf frames     skip after video wakeup, default %d frames\n" \
"  -rl value      minimum average value, default %d (/ 256)\n" \
"  -rt value      maximum average value, default %d (/ 256)\n" \
"  -ru seconds    wakeup interval outside range, default %d s\n" \
"  -rd            draw pixels crossing thresholds on screen, default %s\n" \
"\n" \
"Advanced options:\n" \
"  -d             detach from controlling terminal, default %s\n" \
"  -l logfile     log file if detached, default %s\n" \
"  -si seconds    stats display interval, default %d s\n" \
"  -pi seconds    poll interval, default %.3f s\n" \
"\n"

/*
"Audio options:\n" \
"  -ad device     alsa device, default %s\n" \
"  -ar rate       sample rate, default %d Hz\n" \
"  -ab seconds    sample block, default %g s\n" \
"  -at amplitude  threshold, default %g\n" \
"\n" \

*/

static void config_help()
{
	config c;
	
	config_defaults(&c);
	LOG(HELP_MESSAGE,
		c.host, c.port,
		//c.adev, c.arate, c.ablock, c.athreshold,
		c.vdev, c.vwidth, c.vheight, c.vsleep, c.vwakeup,
		(c.vmask) ? c.vmask : "(none)",
		c.iwidth, c.iheight, c.ipixels, c.iquality,
		c.rspan, c.rithr, c.rpthr,
		c.rbginertia, c.rfginertia, c.rsettle,
		c.ravgmin, c.ravgmax, c.rwakeup,
		(c.rdraw) ? "yes" : "no",
		(c.detach) ? "yes" : "no", c.logfile,
		c.sinterval, c.pinterval);
}

// ***********************************************************************
// 
// ***********************************************************************
static int strtopix(const char *str, char **e_r)
{
	char *e;
	int r;
	
	r = (int)strtol(str, &e, 0);
	if ((*e == 'k') || (*e == 'K')) {
		r <<= 10;
		e++;
	} else if ((*e == 'm') || (*e == 'M')) {
		r <<= 20;
		e++;
	}
	
	if (e_r)
		*e_r = e;
	
	return r;
}

static int strtos(const char *str, char **e_r)
{
	char *e;
	int r;

	r = (int)strtol(str, &e, 0);
	if ((*e == 'h') || (*e == 'H')) {
		r *= 3600;
		e++;
	} else if ((*e == 'm') || (*e == 'M')) {
		r *= 60;
		e++;
	} else if ((*e == 's') || (*e == 'S')) {
		e++;
	}
	
	if (e_r)
		*e_r = e;
	
	return r;
}

// ***********************************************************************
// 
// ***********************************************************************
int config_parse(config *c, int argc, char **argv)
{
	char *a, *s, *e;
	int t;

	config_defaults(c);
	
	for (t = 1; t < argc; t++) {
		a = argv[t];
		
		if ((!strcmp(a, "-h")) || (!strcmp(a, "--help"))) {
			config_help();
			return(1);
		} else if (!strcmp(a, "-d")) {
			c->detach = 1;
			continue;
		} else if (!strcmp(a, "-rd")) {
			c->rdraw = 1;
			continue;
		}
		
		t++;
		
		if (t >= argc) {
			LOG("%s: argument expected\n", a);
			return(-1);
		}
		
		s = argv[t];
		e = NULL;
		
		if (!strcmp(a, "-b")) {
			c->host = s;
		} else if (!strcmp(a, "-p")) {
			c->port = s;
		} else if (!strcmp(a, "-ad")) {
			c->adev = s;
		} else if (!strcmp(a, "-ar")) {
			c->arate = (int)strtol(s, &e, 0);
		} else if (!strcmp(a, "-ab")) {
			c->ablock = strtod(s, &e);
		} else if (!strcmp(a, "-at")) {
			c->athreshold = strtod(s, &e);
		} else if (!strcmp(a, "-vd")) {
			c->vdev = s;
		} else if (!strcmp(a, "-vw")) {
			c->vwidth = strtopix(s, &e);
		} else if (!strcmp(a, "-vh")) {
			c->vheight = strtopix(s, &e);
		} else if (!strcmp(a, "-vs")) {
			c->vsleep = strtos(s, &e);
		} else if (!strcmp(a, "-vu")) {
			c->vwakeup = strtos(s, &e);
		} else if (!strcmp(a, "-vm")) {
			c->vmask = s;
		} else if (!strcmp(a, "-iw")) {
			c->iwidth = strtopix(s, &e);
		} else if (!strcmp(a, "-ih")) {
			c->iheight = strtopix(s, &e);
		} else if (!strcmp(a, "-ip")) {
			c->ipixels = strtopix(s, &e);
		} else if (!strcmp(a, "-rs")) {
			c->rspan = strtos(s, &e);
		} else if (!strcmp(a, "-ri")) {
			c->rithr = strtopix(s, &e);
		} else if (!strcmp(a, "-rp")) {
			c->rpthr = strtopix(s, &e);
		} else if (!strcmp(a, "-r0")) {
			c->rbginertia = strtod(s, &e);
		} else if (!strcmp(a, "-r1")) {
			c->rfginertia = strtod(s, &e);
		} else if (!strcmp(a, "-rf")) {
			c->rsettle = strtopix(s, &e);
		} else if (!strcmp(a, "-rl")) {
			c->ravgmin = strtopix(s, &e);
		} else if (!strcmp(a, "-rt")) {
			c->ravgmax = strtopix(s, &e);
		} else if (!strcmp(a, "-ru")) {
			c->rwakeup = strtos(s, &e);
		} else if (!strcmp(a, "-iq")) {
			c->iquality = (int)strtol(s, &e, 0);
		} else if (!strcmp(a, "-l")) {
			c->logfile = s;
		} else if (!strcmp(a, "-si")) {
			c->sinterval = strtos(s, &e);
		} else if (!strcmp(a, "-pi")) {
			c->pinterval = strtod(s, &e);
		} else {
			LOG("%s: unknown option\n", a);
			return(-1);
		}
		
		if ((e) && ((e == s) || (*e != 0))) {
			LOG("%s: bad value\n", s);
			return(-1);
		}
	}
	
	return(0);
}


