/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdint.h>
#include <stdarg.h>
#include <stdio.h>
#include "csp.h"

#define CSP_YV12	0x32315659
#define CSP_YUYV	0x56595559


// ***********************************************************************
// 
// ***********************************************************************
static void yuyv_to_yuv(unsigned char *yuv, int dw, int dh,
	unsigned char *raw, unsigned char *mask, int w, int h,
	int sx, int sy, int sw, int sh)
{
	unsigned dt, t;
	int dx, dy, x, y;
	int mv;
	
	if (!h)
		return;
	
	for (dy = 0; dy < dh; dy++) {
		y = sy + dy * sh / dh;
		for (dx = 0; dx < dw; dx++) {
			x = sx + dx * sw / dw;
			
			dt = (dy * dw + dx) * 3;
			t = y * w + x;

			mv = mask[t] + 1;

			t *= 2;
			
			yuv[dt + 0] = (raw[t] * mv) >> 8;
			yuv[dt + 1] = raw[(t & (~3)) + 1];
			yuv[dt + 2] = raw[(t & (~3)) + 3];
		}
	}
}

// ***********************************************************************
// 
// ***********************************************************************
int csp_to_yuv(unsigned char *yuv, int dw, int dh,
	void *raw, unsigned char *mask, int width, int height,
	int sx, int sy, int sw, int sh, int fourcc)
{
	if (fourcc == CSP_YUYV) {
		yuyv_to_yuv(yuv, dw, dh, raw, mask, width, height,
			sx, sy, sw, sh);
		return(0);
	}
	
	return(-1);
}


// ***********************************************************************
// 
// ***********************************************************************
void yuv_rect(unsigned char *yuv, int iw, int ih,
	int rx, int ry, int rw, int rh, int cy, int cu, int cv)
{
	int qx, qy, x, y, t;

	qx = rx + rw - 1;
	qy = ry + rh - 1;
	
	if (rx < 0)
		rx = 0;
	if (qx >= iw)
		qx = iw - 1;
	if (rx > qx)
		return;
	
	if (ry < 0)
		ry = 0;
	if (qy >= ih)
		qy = ih - 1;
	if (ry > qy)
		return;
	
	for (y = ry; y <= qy; y++) {
		for (x = rx; x <= qx; x++) {
			t = (y * iw + x) * 3;
			yuv[t + 0] = (unsigned char)cy;
			yuv[t + 1] = (unsigned char)cu;
			yuv[t + 2] = (unsigned char)cv;
		}
	}
}


// ***********************************************************************
// 
// ***********************************************************************
static uint64_t yuv_f8_bitmap[128] = {
0x2a0022002a0000,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x3e3e3e3e3e0000,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x0,0x0,0x0,
0x0,0x8000808080800,0x1414,0x14143e143e1414,
0x81e28281c0a3c08,0x21522408122542,0x2c122a0c12120c,0x808,
0x2010080808081020,0x408101010100804,0x82a1c1c2a0800,0x8083e08080000,
0x810000000000000,0x3e00000000,0x8000000000000,0x2040408101020,
0x1c222a2a221c00,0x10101014181000,0x3c081020241800,0x18242010241800,
0x103e1214181000,0x1824201c043c00,0x1824241c043800,0x4080810203c00,
0x18242418241800,0x1c203824241800,0x8000008000000,0x408000008000000,
0x10080408100000,0x3c003c000000,0x8102010080000,0x10001020241800,
0x3c023a2a323c00,0x42427e42422418,0x3e42423e42423e,0x3c42020202423c,
0x1e22424242221e,0x7e02021e02027e,0x202021e02027e,0x3c42427202423c,
0x4242427e424242,0x1c08080808081c,0x1c222020202030,0x4222120e122242,
0x7e020202020202,0x424242425a6642,0x426252524a4642,0x3c42424242423c,
0x202023e42423e,0x203c52424242423c,0x4222123e42423e,0x3c42403c02423c,
0x808080808083e,0x3c424242424242,0x18182424424242,0x245a4242424242,
0x42422418244242,0x10101010182442,0x7e04080810207e,0x3808080808080838,
0x20101008040402,0x1c1010101010101c,0x442810,0x7e00000000000000,
0x1008,0x5c223c201c0000,0x1e2222221e0202,0x1c2202221c0000,
0x3c2222223c2020,0x3c023e221c0000,0x404041c040438,0x1c22203c221c0000,
0x222222221e0202,0x1c0808080c0008,0xc12101010180010,0x12120e12120202,
0x18080808080808,0x2a2a2a2a160000,0x242424241c0000,0x1c2222221c0000,
0x2021e22221e0000,0x20203c22223c0000,0x404040c340000,0x1e201c023c0000,
0x1c2202020e0202,0x2c121212120000,0x8141422220000,0x142a2a22220000,
0x22140814220000,0x1c22203c22220000,0x3e0408103e0000,0x1008080804080810,
0x808080808080808,0x810101020101008,0x1428,0x0
};

// ***********************************************************************
// 
// ***********************************************************************
static void yuv_f8_putc(unsigned char *yuv, int iw, int ih, int cx, int cy,
	int c)
{
	// available characters
	if ((c < 0) || (c >= 128))
		return;

	// image coordinates
	if ((cx <= -8) || (cy <= -8) || (cx >= iw) || (cy >= ih))
		return;

	uint64_t bitmap = yuv_f8_bitmap[c];
	
	int istride = iw * 3;
	
	if ((cx < 0) || (cy < 0) || (cx + 8 > iw) || (cy + 8 > ih)) {
		int cw = 8;
		int ch = 8;
		
		if (cx >= 0) {
			yuv += cx * 3;
		} else {
			bitmap >>= (-cx);
			cw -= (-cx);
		}
		
		if (cy >= 0) {
			yuv += cy * istride;
		} else {
			bitmap >>= (-cy) * 8;
			ch -= (-cy);
		}
		
		if (cx + 8 > iw)
			cw = iw - cx;
		
		if (cy + 8 > ih)
			ch = ih - cy;

		for (int ty = 0; ty < ch; ty++) {
			uint64_t bitrow = bitmap;
			
			for (int tx = 0; tx < cw; tx++) {
				yuv[tx * 3] = (bitrow & 1) ? 0xff : 0x00;
				bitrow >>= 1;
			}
			
			bitmap >>= 8;
			yuv += istride;
		}
	} else {
		yuv += cy * istride + cx * 3;

		for (int ty = 0; ty < 8; ty++) {
			for (int tx = 0; tx < 8; tx++) {
				yuv[tx * 3] = (bitmap & 1) ? 0xff : 0x00;
				bitmap >>= 1;
			}
			
			yuv += istride;
		}
	}
}

// ***********************************************************************
// 
// ***********************************************************************
static void yuv_f8_puts(unsigned char *yuv, int iw, int ih, int ix, int iy,
	char *str)
{
	int cx, cy;
	int c;
	
	cy = iy;
	cx = ix;
	
	while (1) {
		c = *str;
		
		if (c == 0) {
			break;
		} else if (c == '\n') {
			cx = ix;
			cy += 8;
		} else {
			yuv_f8_putc(yuv, iw, ih, cx, cy, c);
			cx += 8;
		}

		str++;
	}
}

// ***********************************************************************
// 
// ***********************************************************************
void yuv_f8_printf(unsigned char *yuv, int iw, int ih, int ix, int iy,
	char *fmt, ...)
{
	char buff[256];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(buff, sizeof(buff), fmt, ap);
	va_end(ap);

	buff[sizeof(buff) - 1] = 0;
	
	yuv_f8_puts(yuv, iw, ih, ix, iy, buff);
}


// ***********************************************************************
// 
// ***********************************************************************
int yuv_diff(unsigned char *yuv, unsigned char *bkg, unsigned char *frg,
	int bg0, int fg0, int iw, int ih, int thr, int *avg_r, int flags)
{
	uint64_t p, t, cnt;
	int a, s, b, d;
	int bg1 = 256 - bg0;
	int fg1 = 256 - fg0;
	int sum = 0;
	
	p = (uint64_t)iw * (uint64_t)ih;
	cnt = 0;
	
	for (t = 0; t < p; t++) {
		a = yuv[t * 3];
		s = frg[t];
		b = bkg[t];
		
		sum += a;
		
		s = (s * fg0 + a * fg1) >> 8;
		b = (b * bg0 + a * bg1) >> 8;
		
		yuv[t * 3] = (unsigned char)s;
		frg[t] = (unsigned char)s;
		bkg[t] = (unsigned char)b;
		
		d = s - b;
		
		if ((d > thr) || (d < -thr)) {
			if (flags & YUV_DIFF_VISIBLE) {
				yuv[t * 3 + 1] = 0;
				yuv[t * 3 + 2] = 255;
			}
			cnt++;
		}
		
	}
	
	if (avg_r)
		*avg_r = sum / p;
	
	return (int)((cnt << 16) / p);
}
