/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <jpeglib.h>
#include "jpeg_dev.h"

/*
  this code is mainly a copy/paste of parts of the
  libjpeg.doc file shipped along with the 'libjpeg' library
*/

//#ifdef DEBUG_JPEGDEV
#define JDEBUG(s...)	fprintf(stderr, "jpeg_dev: " s)
//#else
//#define JDEBUG(s...)
//#endif

// ********************************************************************
// Huffman evaluator
// ********************************************************************
// Values from ffmpeg libavcodec/mjpeg.c, organized in libjpeg fashion:

static const JHUFF_TBL huff_dc_lum = {
{ /* 0-base */ 0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 }, TRUE };

static const JHUFF_TBL huff_dc_chrom = {
{ /* 0-base */ 0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 },
{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 }, TRUE };

static const JHUFF_TBL huff_ac_lum = {
{ /* 0-base */ 0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d },
{ 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
  0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
  0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
  0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
  0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
  0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
  0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
  0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
  0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
  0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
  0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
  0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
  0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
  0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
  0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
  0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
  0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
  0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
  0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
  0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
  0xf9, 0xfa
}, TRUE };
static const JHUFF_TBL huff_ac_chrom = {
{ /* 0-base */ 0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 },
{ 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
  0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
  0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
  0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
  0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
  0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
  0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
  0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
  0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
  0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
  0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
  0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
  0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
  0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
  0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
  0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
  0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
  0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
  0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
  0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
  0xf9, 0xfa
}, TRUE };

static void huff_dtbl(j_decompress_ptr cinfo, JHUFF_TBL **dst,
	const JHUFF_TBL *src)
{
	// We preserve the original table if it exists
	if (*dst == NULL) {
		*dst = jpeg_alloc_huff_table((j_common_ptr)cinfo);
		memcpy(*dst, src, sizeof(JHUFF_TBL));
	}
}

static void huff_ctbl(j_compress_ptr cinfo, JHUFF_TBL **dst,
	const JHUFF_TBL *src)
{
	if (*dst == NULL) *dst = jpeg_alloc_huff_table((j_common_ptr)cinfo);

	// We override any existing table
	memcpy(*dst, src, sizeof(JHUFF_TBL));
}



// ********************************************************************
// Decompression
// ********************************************************************

// **************************************************
struct custom_src_t {
	struct jpeg_source_mgr jsrc;
	void *buff;
	int size;
	int used_size;
};

static void jpegdev_init_source(j_decompress_ptr cinfo)
{
	struct custom_src_t *s = (struct custom_src_t *)cinfo->src;

	s->jsrc.next_input_byte = s->buff;
	s->jsrc.bytes_in_buffer = s->size;
}

static boolean jpegdev_fill_input_buffer(j_decompress_ptr cinfo)
{
	JDEBUG("Incomplete file, filling with random data\n");
	jpegdev_init_source(cinfo);
	return(TRUE);
}

static void jpegdev_skip_input_data(j_decompress_ptr cinfo, long n)
{
	struct custom_src_t *s = (struct custom_src_t *)cinfo->src;

	if (n <= 0) return;
	if (n > (long)s->jsrc.bytes_in_buffer) {
		jpegdev_fill_input_buffer(cinfo);
		return;
	}
	s->jsrc.bytes_in_buffer -= n;
	s->jsrc.next_input_byte += n;
}

static void jpegdev_term_source(j_decompress_ptr cinfo)
{
	struct custom_src_t *s = (struct custom_src_t *)cinfo->src;
	
	s->used_size = s->size - s->jsrc.bytes_in_buffer;
}

// **************************************************
int jpegdev_extract(void *ddata, void *cdata, int csize,
	int width, int height, int flags)
{
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct custom_src_t custom_src;
	JSAMPROW row_pointer[1];
	int row_stride;

	while( (csize > 4) && (((unsigned short *)cdata)[0] != 0xd8ff) ) {
		cdata++;
		csize--;
	}
	
	if (csize <= 4) return(-1);
	
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);

	// stdio: jpeg_stdio_src(&cinfo, FILE *)
	custom_src.buff = cdata;
	custom_src.size = csize;
		// custom_src.jsrc.next_input_byte = cdata;
		// custom_src.jsrc.bytes_in_buffer = csize;
	custom_src.jsrc.init_source = jpegdev_init_source;
	custom_src.jsrc.fill_input_buffer = jpegdev_fill_input_buffer;
	custom_src.jsrc.skip_input_data = jpegdev_skip_input_data;
	custom_src.jsrc.resync_to_restart = jpeg_resync_to_restart;
	custom_src.jsrc.term_source = jpegdev_term_source;
	cinfo.src = &custom_src.jsrc;

	jpeg_read_header(&cinfo, TRUE);
	// returns:
	// JPEG_HEADER_OK
	// JPEG_HEADER_TABLES_ONLY (only if require_image=FALSE was passed)
	// JPEG_SUSPENDED (only if the source manager suspended IO)

	if (((int)cinfo.image_width != width)
	||  ((int)cinfo.image_height != height)) {
		jpeg_destroy_decompress(&cinfo);
		return(-1);
	}

	huff_dtbl(&cinfo, &cinfo.dc_huff_tbl_ptrs[0], &huff_dc_lum);
	huff_dtbl(&cinfo, &cinfo.dc_huff_tbl_ptrs[1], &huff_dc_chrom);
	huff_dtbl(&cinfo, &cinfo.ac_huff_tbl_ptrs[0], &huff_ac_lum);
	huff_dtbl(&cinfo, &cinfo.ac_huff_tbl_ptrs[1], &huff_ac_chrom);

	switch(flags & JPEGDEV_COLORSPACE_MASK) {
	case JPEGDEV_PACKED_YUV:
		cinfo.out_color_space = JCS_YCbCr;
		cinfo.output_components = cinfo.out_color_components = 3;
		break;
	case JPEGDEV_GRAYSCALE:
		cinfo.out_color_space = JCS_GRAYSCALE;
		cinfo.output_components = cinfo.out_color_components = 1;
		break;
	case JPEGDEV_PACKED_RGB:
	default:
		cinfo.out_color_space = JCS_RGB;
		cinfo.output_components = cinfo.out_color_components = 3;
		break;
	}
	row_stride = cinfo.image_width * cinfo.output_components;

	jpeg_start_decompress(&cinfo);

	while (cinfo.output_scanline < cinfo.image_height) {
		switch(flags & JPEGDEV_ROW_ORDER_MASK) {
		case JPEGDEV_TOP_DOWN:
			row_pointer[0] =
				ddata + (cinfo.output_scanline * row_stride);
			break;
		case JPEGDEV_BOTTOM_UP:
			row_pointer[0] =
				ddata + (row_stride * (cinfo.image_height
					- cinfo.output_scanline - 1));
			break;
		}
		jpeg_read_scanlines(&cinfo, row_pointer, 1);
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

	return(custom_src.used_size);
}

// **************************************************
int jpegdev_getsize(void *cdata, int csize, int *w, int *h)
{
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct custom_src_t custom_src;

	while( (csize > 4) && (((unsigned short *)cdata)[0] != 0xd8ff) ) {
		cdata++;
		csize--;
	}
	
	if (csize <= 4) return(-1);
	
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);

	// stdio: jpeg_stdio_src(&cinfo, FILE *)
	custom_src.buff = cdata;
	custom_src.size = csize;
		// custom_src.jsrc.next_input_byte = cdata;
		// custom_src.jsrc.bytes_in_buffer = csize;
	custom_src.jsrc.init_source = jpegdev_init_source;
	custom_src.jsrc.fill_input_buffer = jpegdev_fill_input_buffer;
	custom_src.jsrc.skip_input_data = jpegdev_skip_input_data;
	custom_src.jsrc.resync_to_restart = jpeg_resync_to_restart;
	custom_src.jsrc.term_source = jpegdev_term_source;
	cinfo.src = &custom_src.jsrc;

	jpeg_read_header(&cinfo, TRUE);

	if (w) *w = cinfo.image_width;
	if (h) *h = cinfo.image_height;
	jpeg_destroy_decompress(&cinfo);

	return(0);
}



// ********************************************************************
// Compression
// ********************************************************************

// **************************************************
struct custom_dst_t {
	struct jpeg_destination_mgr jdst;
	void *buff;
	int size;
	int used_size;
};

static void jpegdev_init_destination(j_compress_ptr cinfo)
{
	struct custom_dst_t *d = (struct custom_dst_t *)cinfo->dest;

	d->jdst.next_output_byte = d->buff;
	d->jdst.free_in_buffer = d->size;
}

static boolean jpegdev_empty_output_buffer(j_compress_ptr cinfo)
{
	JDEBUG("empty_output_buffer() called with memory destination\n");
	JDEBUG("buffer size %d too small for output (space left: %d)\n",
		((struct custom_dst_t *)cinfo->dest)->size,
		(int)cinfo->dest->free_in_buffer);
	jpegdev_init_destination(cinfo);

	return(TRUE);
}

static void jpegdev_term_destination(j_compress_ptr cinfo)
{
	struct custom_dst_t *d = (struct custom_dst_t *)cinfo->dest;

	d->used_size = d->size - d->jdst.free_in_buffer;
}


// **************************************************
int jpegdev_compress(void *cdata, int maxsize,
	void *ddata, int width, int height, int flags)
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct custom_dst_t custom_dst;
	JSAMPROW row_pointer[1];
	int row_stride;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);

	// stdio: jpeg_stdio_dest(&cinfo, FILE *);
	custom_dst.buff = cdata;
	custom_dst.size = maxsize;
	custom_dst.used_size = 0;
		// custom_dst.jdst.next_output_byte = cdata;
		// custom_dst.jdst.free_in_buffer = maxsize;
	custom_dst.jdst.init_destination = jpegdev_init_destination;
	custom_dst.jdst.empty_output_buffer = jpegdev_empty_output_buffer;
	custom_dst.jdst.term_destination = jpegdev_term_destination;
	cinfo.dest = &custom_dst.jdst;

	cinfo.image_width = width;
	cinfo.image_height = height;
	switch(flags & JPEGDEV_COLORSPACE_MASK) {
	case JPEGDEV_PACKED_YUV:
		cinfo.in_color_space = JCS_YCbCr;
		cinfo.input_components = 3;
		break;
	case JPEGDEV_GRAYSCALE:
		cinfo.in_color_space = JCS_GRAYSCALE;
		cinfo.input_components = 1;
		break;
	case JPEGDEV_PACKED_RGB:
	default:
		cinfo.in_color_space = JCS_RGB;
		cinfo.input_components = 3;
		break;
	}
	row_stride = cinfo.image_width * cinfo.input_components;

	jpeg_set_defaults(&cinfo);
	if ((flags & JPEGDEV_QUALITY_MASK) == JPEGDEV_QUALITY_CUSTOM) {
		jpeg_set_quality(&cinfo,
			(flags >> JPEGDEV_CUSTOM_Q_SHIFT) & 0xff, FALSE);
	}

	switch(flags & JPEGDEV_TABLES_MASK) {
	case JPEGDEV_MJPEG_TABLES:
		huff_ctbl(&cinfo, &cinfo.dc_huff_tbl_ptrs[0], &huff_dc_lum);
		huff_ctbl(&cinfo, &cinfo.dc_huff_tbl_ptrs[1], &huff_dc_chrom);
		huff_ctbl(&cinfo, &cinfo.ac_huff_tbl_ptrs[0], &huff_ac_lum);
		huff_ctbl(&cinfo, &cinfo.ac_huff_tbl_ptrs[1], &huff_ac_chrom);
		jpeg_start_compress(&cinfo, FALSE);
		break;
	case JPEGDEV_INTERCHANGE:
	default:
		jpeg_start_compress(&cinfo, TRUE);
		break;
	}

	while (cinfo.next_scanline < cinfo.image_height) {
		switch(flags & JPEGDEV_ROW_ORDER_MASK) {
		case JPEGDEV_TOP_DOWN:
			row_pointer[0] =
				ddata + (cinfo.next_scanline * row_stride);
			break;
		case JPEGDEV_BOTTOM_UP:
			row_pointer[0] =
				ddata + (row_stride * (cinfo.image_height
					- cinfo.next_scanline - 1));
			break;
		}
		jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	return(custom_dst.used_size);
}


// ********************************************************************
// To-file and from-file utilities
// ********************************************************************

// **************************************************
int jpegdev_compress_file(char *fname, void *ddata, int width, int height,
	int flags)
{
	void *cdata;
	int sz, fd, o, n;

	// rule of the thumb for maximum compressed size
	switch(flags & JPEGDEV_COLORSPACE_MASK) {
	case JPEGDEV_GRAYSCALE:
		sz = 1024 + width * height;
		break;
	case JPEGDEV_PACKED_YUV:
	case JPEGDEV_PACKED_RGB:
	default:
		sz = 1024 + width * height * 3;
		break;
	}

	cdata = malloc(sz);
	if (!cdata) {
		JDEBUG("%s: malloc() failed\n", fname);
		return(-1);
	}

	switch(flags & JPEGDEV_OPEN_MODE_MASK) {
	case JPEGDEV_APPEND_FILE:
		fd = open(fname, O_CREAT | O_WRONLY | O_APPEND, 0644);
		break;
	case JPEGDEV_CREATE_FILE:
	default:
		fd = creat(fname, 0644);
		break;
	}
	
	if (fd < 0) {
		JDEBUG("%s: %s\n", fname, strerror(errno));
		free(cdata);
		return(-1);
	}

	sz = jpegdev_compress(cdata, sz, ddata, width, height, flags);

	o = 0;
	while(o < sz) {
		n = write(fd, cdata + o, sz - o);
		if ((n < 0) && ((errno == EINTR) || (errno == EAGAIN))) n = 0;
		else if (n < 0) {
			JDEBUG("%s: %s\n", fname, strerror(errno));
			free(cdata);
			close(fd);
			return(-1);
		}
			
		o += n;
	}

	free(cdata);
	close(fd);
	return(sz);
}


// **************************************************
void *jpegdev_extract_file(char *fname, int *w, int *h, int flags)
{
	struct stat st;
	int fw, fh;
	int fd, n, o;
	void *cdata;
	void *ddata;

	fd = open(fname, O_RDONLY);
	if (fd < 0) {
		JDEBUG("%s: %s\n", fname, strerror(errno));
		return(NULL);
	}

	if (fstat(fd, &st)) {
		JDEBUG("%s: stat(): %s\n", fname, strerror(errno));
		close(fd);
		return(NULL);
	}

	cdata = malloc(st.st_size);
	if (!cdata) {
		JDEBUG("%s: malloc(%d) failed\n", fname, (int)st.st_size);
		close(fd);
		return(NULL);
	}

	o = 0;
	while (o < st.st_size) {
		n = read(fd, cdata + o, st.st_size - o);
		if ((n < 0) && ((errno == EINTR) || (errno == EAGAIN))) n = 0;
		else if (n <= 0) {
			JDEBUG("%s: %s\n", fname, strerror(errno));
			free(cdata);
			close(fd);
			return(NULL);
		}
		o += n;
	}

	close(fd);

	if (jpegdev_getsize(cdata, st.st_size, &fw, &fh) < 0) {
		free(cdata);
		return(NULL);
	}

	switch(flags & JPEGDEV_COLORSPACE_MASK) {
	case JPEGDEV_GRAYSCALE:
		ddata = malloc(fw * fh);
		break;
	case JPEGDEV_PACKED_YUV:
	case JPEGDEV_PACKED_RGB:
	default:
		ddata = malloc(fw * fh * 3);
		break;
	}

	n = jpegdev_extract(ddata, cdata, st.st_size, fw, fh, flags);

	free(cdata);

	if (n < 0) {
		free(ddata);
		return(NULL);
	}

	if (n != st.st_size) {
		JDEBUG("%s: trailing garbage at end of file (%d bytes)\n",
			fname, (int)st.st_size - n);
	}

	if (w) *w = fw;
	if (h) *h = fh;
	return(ddata);
}


