/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include "jpeg_dev.h"
#include "log.h"
#include "mask.h"

// ***********************************************************************
// 
// ***********************************************************************
static unsigned char mask_default[1] = { 255 };

int mask_init(mask_data *m, char *path)
{
	if (path == NULL) {
		m->od = mask_default;
		m->ow = 1;
		m->oh = 1;
	} else {
		m->od = jpegdev_extract_file(path, &m->ow, &m->oh,
			JPEGDEV_GRAYSCALE);

		if (!m->od)
			return(-1);
	}
	
	m->cw = 0;
	m->ch = 0;
	m->cd = NULL;
	
	m->mx = 0;
	m->my = 0;
	m->mw = 0;
	m->mh = 0;
	return(0);
}

// ***********************************************************************
// 
// ***********************************************************************
void mask_clear(mask_data *m)
{
	if ((m->od) && (m->od != mask_default))
		free(m->od);
	
	if (m->cd)
		free(m->cd);
}

// ***********************************************************************
// 
// ***********************************************************************
int mask_setup(mask_data *m, int dw, int dh)
{
	int dx, dy, sxa, sxb, sxf, sya, syb, syf, sw, sh;
	int x_min, x_max, y_min, y_max, v, lv;
	unsigned char *dd, *sd;
	
	if ((dw == m->cw) && (dh == m->ch))
		return(0);
	
	if (m->cd)
		free(m->cd);
	
	m->cw = 0;
	m->ch = 0;
	m->cd = NULL;
	
	dd = malloc(dw * dh * sizeof(unsigned char));
	
	if (!dd)
		return(-1);

	// scale mask to fit video device resolution,
	// and detect bounding box
	sw = m->ow;
	sh = m->oh;
	sd = m->od;
	
	x_min = dw;
	x_max = 0;
	y_min = dh;
	y_max = 0;
	
	for (dy = 0; dy < dh; dy++) {
		syf = (dy * sh << 8) / dh;
		sya = syf >> 8;
		syb = sya + 1;
		if (syb == sh)
			syb = sya;
		syf &= 255;
		
		lv = 0;
		
		for (dx = 0; dx < dw; dx++) {
			sxf = (dx * sw << 8) / dw;
			sxa = sxf >> 8;
			sxb = sxa + 1;
			if (sxb == sw)
				sxb = sxa;
			sxf &= 255;
			
			v = sd[sya * sw + sxa] * (256 - syf) * (256 - sxf)
			  + sd[sya * sw + sxb] * (256 - syf) * sxf
			  + sd[syb * sw + sxa] * syf * (256 - sxf)
			  + sd[syb * sw + sxb] * syf * sxf;
			v >>= 16;
			
			if (v >= MASK_THRESHOLD) {
				lv = 1;
				
				if (dx > x_max)
					x_max = dx;
				if (dx < x_min)
					x_min = dx;
			}

			dd[dy * dw + dx] = v;
		}
		
		if (lv) {
			if (dy > y_max)
				y_max = dy;
			if (dy < y_min)
				y_min = dy;
		}
	}
	
	if ((x_max < x_min) || (y_max < y_min)) {
		log_time("mask_setup(): empty (black) image\n");
		return(-1);
	}

	m->mx = x_min;
	m->my = y_min;
	m->mw = x_max - x_min + 1;
	m->mh = y_max - y_min + 1;

	m->cw = dw;
	m->ch = dh;
	m->cd = dd;

	return(0);
}


