/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#define _POSIX_C_SOURCE 201605L
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "log.h"
#include "network.h"

#define NETSERV_BACKPE	10

#ifndef SOCK_CLOEXEC
#define SOCK_CLOEXEC	0
#endif

// **************************************************************************
// 
// **************************************************************************
int  net_init(const char *host, const char *service)
{
	char r_service[32];
	char r_host[64];
	struct addrinfo hints;
	struct addrinfo *res, *rp;
	int fd, e;
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG;
	
	e = getaddrinfo(host, service, &hints, &res);
	
	if (e != 0) {
		log_time("net_init(): getaddrinfo(%s, %s): %s\n",
			host, service, gai_strerror(e));
		if (e == EAI_SYSTEM)
			log_time("\t(%s)\n", strerror(errno));
		return(-1);
	}

	for (rp = res; rp != NULL; rp = rp->ai_next) {
		e = getnameinfo(rp->ai_addr, rp->ai_addrlen,
			r_host, sizeof(r_host), r_service, sizeof(r_service),
			NI_NUMERICHOST | NI_NUMERICSERV);
		
		if (e) {
			log_time("net_init(): getnameinfo(): %s\n",
				gai_strerror(e));
			if (e == EAI_SYSTEM)
				log_time("\t(%s)\n", strerror(errno));
		} else {
			log_time("%s:%s resolved to "
				"%s:%s\n", host, service, r_host, r_service);
		}
		
		fd = socket(rp->ai_family, rp->ai_socktype | SOCK_CLOEXEC,
			rp->ai_protocol);
		
		if (fd == -1) {
			log_time("net_init(): socket(): %s\n",
				strerror(errno));
			continue;
		}
	
		if (bind(fd, rp->ai_addr, rp->ai_addrlen)) {
			log_time("net_init(): bind(): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		if (listen(fd, NETSERV_BACKPE)) {
			log_time("net_init(): listen(): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		if (fcntl(fd, F_SETFL, O_NONBLOCK)) {
			log_time("net_init(): "
				"fcntl(F_SETFL, O_NONBLOCK): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		break;
	}
	
	freeaddrinfo(res);
	
	if (!rp)
		return(-1);
	
	return(fd);
}

// **************************************************************************
// 
// **************************************************************************
int net_accept(int sfd, net_addr *peer)
{
	struct sockaddr_storage pa;
	socklen_t pa_len;
	int cfd, e;
	
	pa_len = sizeof(pa);
	cfd = accept(sfd, (struct sockaddr *)&pa, &pa_len);
	
	if (cfd == -1) {
		log_time("net_accept(): accept(): %s\n",
			strerror(errno));
		return(-1);
	}
	
	if (fcntl(cfd, F_SETFL, O_NONBLOCK)) {
		log_time("net_accept(): "
			"fcntl(F_SETFL, O_NONBLOCK): %s\n",
			strerror(errno));
		close(cfd);
		return(-1);
	}

	if (peer) {
		e = getnameinfo((struct sockaddr *)&pa, pa_len,
			peer->host, sizeof(peer->host),
			peer->service, sizeof(peer->service), NI_NUMERICSERV);
		
		if (e) {
			log_time("net_accept(): getnameinfo(): %s\n",
				gai_strerror(e));
			if (e == EAI_SYSTEM)
				log_time("\t(%s)\n", strerror(errno));
			close(cfd);
			return(-1);
		}
	}
	
	return(cfd);
}




