/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include "log.h"
#include "csp.h"
#include "roll.h"

// ***********************************************************************
// 
// ***********************************************************************
void roll_init(roll *r, config *c)
{
	memset(r->history, 0, ROLL_HISTORY * sizeof(int));
	r->ptr = 0;
	r->origin = 0;
	r->span = c->rspan;
}

void roll_clear(roll *r)
{
	(void)r;
}

// ***********************************************************************
// 
// ***********************************************************************
void roll_push(roll *r, int val, time_t time)
{
	int loc;
	
	if (r->origin == 0)
		r->origin = time - r->span;
	
	loc = (int)((time - r->origin) * ROLL_HISTORY / r->span);
	
	while (r->ptr < loc) {
		r->ptr++;
		r->history[r->ptr % ROLL_HISTORY] = -1;
	}
	
	if (r->history[r->ptr % ROLL_HISTORY] < 0)
		r->history[r->ptr % ROLL_HISTORY] = val;
	else
		r->history[r->ptr % ROLL_HISTORY] += val;
}

// ***********************************************************************
// 
// ***********************************************************************
#define TIME_UNITS	((int)(sizeof(time_unit) / sizeof(time_unit[0])))
static time_t time_unit[] = {
	60, 300, 600, 1800, 3600, 7200, 10800, 14400,
	21600, 28800, 43200, 86400, 604800
};

static time_t roll_time_unit(int px, int iw, time_t span)
{
	int j, width;
	
	for (j = 0; j < TIME_UNITS - 1; j++) {
		width = (int)time_unit[j] * iw / (int)span;
		
		if (width > px)
			break;
	}
	
	return time_unit[j];
}

// ***********************************************************************
// 
// ***********************************************************************
void roll_plot(roll *r, unsigned char *yuv, int iw, int ih)
{
	struct tm *t;
	time_t now, time, u;
	int block, loc, loc0, val, vmax;
	int x0, x1, rh, lh;
	
	rh = 64;
	
	// find highest bar value
	vmax = 0;
	for (loc = 0; loc < ROLL_HISTORY; loc++) {
		if (r->history[loc] > vmax)
			vmax = r->history[loc];
	}
	
	if (vmax == 0)
		return;
	
	// display bars
	loc0 = r->ptr - ROLL_HISTORY + 1;
	
	for (block = 0; block < ROLL_HISTORY; block++) {
		loc = loc0 + block;
		//time = r->origin + (loc * r->span / ROLL_HISTORY);
		val = r->history[loc % ROLL_HISTORY];
		
		x0 = (block + 0) * iw / ROLL_HISTORY;
		x1 = (block + 1) * iw / ROLL_HISTORY;
		
		if (val >= 0) {
			lh = rh * val / vmax;
		
			yuv_rect(yuv, iw, ih, x0, ih - lh - 2, x1 - x0, lh,
				128, 255, 0);
			yuv_rect(yuv, iw, ih, x0, ih - 2, x1 - x0, 2,
				128, 0, 255);
		}
	}
	
	// current time
	now = r->origin + r->ptr * r->span / ROLL_HISTORY;

	// display ticks
	u = roll_time_unit(30, iw, r->span);
	time = now - (now % u);
	
	while (1) {
		loc = (int)(time - r->origin) * ROLL_HISTORY / r->span;
		block = loc - loc0;
		x0 = block * iw / ROLL_HISTORY;
		
		if (x0 < 0)
			break;
		
		yuv_rect(yuv, iw, ih, x0, ih - 12, 1, 4, 255, 128, 128);

		time -= u;
	}

	// display time
	u = roll_time_unit(80, iw, r->span);
	time = now - (now % u);
	
	while (1) {
		loc = (int)(time - r->origin) * ROLL_HISTORY / r->span;
		block = loc - loc0;
		x0 = block * iw / ROLL_HISTORY;
		
		if (x0 < -40)
			break;
		
		t = localtime(&time);
		
		yuv_rect(yuv, iw, ih, x0, ih - 12, 1, 12, 255, 128, 128);
		
		//x1 = (x0 + 40 < iw) ? x0 + 1 : x0 - 40;
		x1 = x0 + 1;
		
		yuv_f8_printf(yuv, iw, ih, x1, ih - 8, "%02d:%02d",
			t->tm_hour, t->tm_min);

		time -= u;
	}
	
}
