/*
    This file is part of l_mon.
    Copyright (C) 2016-2022  Laurent Poirrier

    l_mon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with l_mon.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <poll.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "log.h"
#include "video_dev.h"
#include "jpeg_dev.h"
#include "hutils.h"
#include "files.h"
#include "csp.h"
#include "refbuf.h"
#include "server.h"

// ***********************************************************************
// 
// ***********************************************************************
static int frame_handle(server *s)
{
	struct tm *t;
	void *raw;
	int diff, motion, settled, inrange;
	int avg;
	
	raw = video_capture(&s->fdev);
	
	if (!raw)
		return(-1);
	
	if (csp_to_yuv(s->yuv_buff, s->iw, s->ih,
			raw, s->fmask.cd, s->fdev.width, s->fdev.height,
			s->fmask.mx, s->fmask.my, s->fmask.mw, s->fmask.mh,
			s->fdev.fourcc)) {
		log_time("Capture pixel format (fourcc) not supported\n");
		return(-1);
	}
	
	if (video_release(&s->fdev, raw))
		return(-1);
	
	if (s->conf->rithr >= 1) {
		diff = yuv_diff(s->yuv_buff, s->bkg_buff, s->frg_buff,
			s->conf->rbginertia * 256.0,
			s->conf->rfginertia * 256.0,
			s->iw, s->ih, s->conf->rithr, &avg,
			s->conf->rdraw ? YUV_DIFF_VISIBLE : 0);
	} else {
		diff = 0;
		avg = 0;
	}
	
	inrange = ((avg >= s->conf->ravgmin) && (avg <= s->conf->ravgmax));
	motion = (diff > s->conf->rpthr);
	settled = ((s->conf->rsettle < 1)
		|| (s->recv_frames >= s->switch_frame + s->conf->rsettle));

	yuv_rect(s->yuv_buff, s->iw, s->ih,
		((s->recv_frames & 127) - 16) * (s->iw * 9) / 8 / 128,
		0,
		s->iw / 8, 8,
		128, 255, 0);

	if (s->conf->rdraw) {
		yuv_f8_printf(s->yuv_buff, s->iw, s->ih, s->iw - 120, 0,
			" %3d%c %7d%c%c",
			avg,
			inrange ? ' ' : 'R',
			diff,
			settled ? ' ' : 'X',
			motion ? '!' : ' ');
	}
	
	t = localtime(&s->curtime);
	
	yuv_f8_printf(s->yuv_buff, s->iw, s->ih, 0, 0,
		"%04d-%02d-%02d %02d:%02d:%02d",
		t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
		t->tm_hour, t->tm_min, t->tm_sec);

	if (inrange && motion && settled) {
		roll_push(&s->pixroll, 1, s->curtime);
		
		// keep awake
		s->fatime = s->curtime;
	} else {
		roll_push(&s->pixroll, 0, s->curtime);
	}
	
	roll_plot(&s->pixroll, s->yuv_buff, s->iw, s->ih);

	s->wakeup_delay = (inrange) ? s->conf->vwakeup : s->conf->rwakeup;

	s->fready = 1;
	
	if (s->fcur)
		refbuf_ref(s->fcur, -1);
	s->fcur = NULL;

	//log_time("  received frame %d\n", s->recv_frames);
	s->recv_frames++;

	return(0);
}

// ***********************************************************************
// 
// ***********************************************************************
static int frame_compute(server *s)
{
	size_t jsize;
	refbuf *b;

	jsize = (size_t)jpegdev_compress(s->jpeg_buff,
		(int)s->jpeg_buff_size,
		s->yuv_buff, s->iw, s->ih,
		JPEGDEV_TOP_DOWN | JPEGDEV_PACKED_YUV
		| JPEGDEV_QUALITY(s->conf->iquality));
	
	b = refbuf_create(&s->fpool, jsize);
	if (!b)
		return(-1);
	
	memcpy(b->data, s->jpeg_buff, jsize);
	
	if (s->fcur)
		refbuf_ref(s->fcur, -1);
	s->fcur = b;
	
	//log_time("  computed frame %d\n", s->fcur->seq);
	
	return(0);
}

// ***********************************************************************
// 
// ***********************************************************************
static client *client_create(server *s)
{
	client *c;
	
	c = malloc(sizeof(client));
	if (!c) {
		log_time("client_create(): malloc(): %s\n", strerror(errno));
		return(NULL);
	}
	
	c->fd = net_accept(s->d.fd, &c->addr);

	if (c->fd < 0) {
		log_time("accept(): %s\n", strerror(errno));
		free(c);
		return(NULL);
	}
	
	
	//log_time("Connection from %s:%s\n", c->addr.host, c->addr.service);

	c->flags = 0;
	c->pos = 0;
	c->hsize = c->hpos = 0;
	c->fbuf = NULL;
	c->fpos = 0;
	c->fseq = -1;
	
	c->prev = s->d.prev;
	c->next = &s->d;
	s->d.prev->next = c;
	s->d.prev = c;
	s->nd++;
	return(c);
}

static void client_destroy(server *s, client *c)
{
	close(c->fd);
	
	if (c->fbuf)
		refbuf_ref(c->fbuf, -1);
	
	c->prev->next = c->next;
	c->next->prev = c->prev;
	free(c);
	s->nd--;
}

// ***********************************************************************
// 
// ***********************************************************************
static void client_input(server *s, client *c)
{
	refbuf *fb;
	const char *cdata, *cname, *ctype;
	size_t hsize, csize, urlsize;
	
	// Input buffer overflow
	if (c->pos == REQ_SIZE - 1) {
		c->flags |= CLIENT_KILL;
		return;
	}
	
	// Test request completion
	if (!strstr(c->buff, "\r\n\r\n"))
		return;
	
	// We are not able to handle two simultaneous requests
	if ((c->hsize) || (c->fbuf)) {
		c->flags |= CLIENT_KILL;
		return;
	}
	
	// Reset input buffer
	c->pos = 0;

	//log_time("Request:\n%s---\n", c->buff);

	// Get URL
	urlsize = hget_url(s->url, REQ_SIZE, c->buff);

	// Request not recognized
	if (urlsize == 0) {
		c->flags |= CLIENT_KILL;
		return;
	}
	
	// Frame
	if (strends(s->url, "/frame.jpg")) {
		c->flags |= CLIENT_E_FRAME;
		//log_time("  %s:%s frame.jpg\n",
		//	c->addr.host, c->addr.service);
		return;
	}
	
	// Frame
	if (strends(s->url,  "/video.mjpg")) {
		c->flags |= CLIENT_E_VIDEO;
		//log_time("  %s:%s video.mjpg\n",
		//	c->addr.host, c->addr.service);
		
		hsize = hheader_multi_start(c->hbuf, HEADER_SIZE, s->curtime,
			HHEADER_NOCACHE);
		
		if (hsize > HEADER_SIZE)
			hsize = HEADER_SIZE;
		
		c->hsize = hsize;
		c->hpos = 0;
		return;
	}
	
	// Detect static files
	cdata = file_data(&cname, &ctype, &csize, s->url);
	
	// Send redirect
	if (cdata == NULL) {
		if (urlsize > HEADER_SIZE / 4) {
			urlsize = HEADER_SIZE / 4;
			s->url[urlsize] = 0;
		}
		
		log_time("  %s:%s '%s' (moved)\n",
			c->addr.host, c->addr.service, s->url);

		hsize = hheader_move(c->hbuf, HEADER_SIZE, s->curtime, s->url);
		
		if (hsize > HEADER_SIZE)
			hsize = HEADER_SIZE;
		
		c->hsize = hsize;
		c->hpos = 0;
		
		c->fbuf = NULL;
		c->fpos = 0;
		c->flags |= CLIENT_LAST_MSG;
		return;
	}
	
	// Send static files
	log_time("  %s:%s %s (%s | %zd bytes)\n",
		c->addr.host, c->addr.service, cname, ctype, csize);
	
	hsize = hheader_ok(c->hbuf, HEADER_SIZE, s->curtime,
		ctype, csize, 0);
	
	if (hsize > HEADER_SIZE)
		hsize = HEADER_SIZE;
	
	c->hsize = hsize;
	c->hpos = 0;
	
	fb = refbuf_create(&s->dpool, csize);

	if (!fb) {
		c->flags |= CLIENT_KILL;
		return;
	}

	memcpy(fb->data, cdata, csize);

	c->fbuf = fb;
	c->fpos = 0;
	c->flags |= CLIENT_LAST_MSG;
	return;
}

// ***********************************************************************
// 
// ***********************************************************************
static void client_handle(server *s, client *c)
{
	size_t hsize;
	
	if (c->flags & CLIENT_IN) {
		client_input(s, c);
	}
	
	if ((c->hsize) && (c->hpos == c->hsize)) {
		s->sent_bytes += c->hsize;
		c->hsize = 0;
		c->hpos = 0;

		if ((!c->fbuf) && (c->flags & CLIENT_LAST_MSG))
			c->flags |= CLIENT_KILL;
	}
	
	if ((c->fbuf) && (c->fpos == c->fbuf->size)) {
		s->sent_bytes += c->fbuf->size;
		
		refbuf_ref(c->fbuf, -1);
		c->fbuf = NULL;
		c->fpos = 0;
		
		if (c->flags & CLIENT_LAST_MSG)
			c->flags |= CLIENT_KILL;
		
		//log_time("%s:%s: buffer sent\n",
		//	c->addr.host, c->addr.service);
	}
	
	if (c->flags & (CLIENT_E_FRAME | CLIENT_E_VIDEO)) {
		if (!s->fready) {
			s->fatime = s->curtime;
		} else if (!s->fcur) {
			frame_compute(s);
		}
	}
	
	
	if ((s->fcur) && (c->flags & CLIENT_E_FRAME)) {
		s->fatime = s->curtime;
		
		refbuf_ref(s->fcur, 1);

		ASSERT(c->hsize == 0);
		ASSERT(c->fbuf == NULL);

		hsize = hheader_ok(c->hbuf, HEADER_SIZE, s->curtime,
			"image/jpeg", s->fcur->size, HHEADER_NOCACHE);
		
		if (hsize > HEADER_SIZE)
			hsize = HEADER_SIZE;
		
		c->hsize = hsize;
		c->hpos = 0;
		
		c->fbuf = s->fcur;
		c->fpos = 0;
		c->fseq = c->fbuf->seq;
		
		c->flags &= ~CLIENT_E_FRAME;
		c->flags |= CLIENT_LAST_MSG;
		
		//log_time("  %s:%s sent frame #%d seq %d\n",
		//	c->addr.host, c->addr.service,
		//	s->sent_frames, c->fseq);

		s->sent_frames++;
	}
	
	if ((s->fcur) && (c->flags & CLIENT_E_VIDEO)
	&&  (s->fcur->seq != c->fseq)
	&&  (c->hsize == 0) && (c->fbuf == NULL)) {
		s->fatime = s->curtime;
		
		refbuf_ref(s->fcur, 1);

		hsize = hheader_multi_part(c->hbuf, HEADER_SIZE, s->curtime,
			"image/jpeg", s->fcur->size, HHEADER_NOCACHE);
		
		if (hsize > HEADER_SIZE)
			hsize = HEADER_SIZE;
		
		c->hsize = hsize;
		c->hpos = 0;

		c->fbuf = s->fcur;
		c->fseq = c->fbuf->seq;
		
		s->sent_frames++;
		//log_time("  %s:%s sent video frame %d\n",
		//	c->addr.host, c->addr.service, c->fseq);
	}

	if (c->flags & CLIENT_KILL) {
		client_destroy(s, c);
	}
}

// ***********************************************************************
// 
// ***********************************************************************
static int server_init(server *s, config *c)
{
	s->conf = c;
	s->curtime = time(NULL);
	
	s->fdev.fd = -1;
	
	if (mask_init(&s->fmask, c->vmask))
		return(-1);
	
	pool_init(&s->fpool);
	s->fcur = NULL;
	s->fready = 0;
	s->fatime = s->curtime;
	s->iw = 0;
	s->ih = 0;
	
	s->d.fd = net_init(c->host, c->port);
	
	if (s->d.fd < 0) {
		pool_clear(&s->fpool);
		return(-1);
	}
	
	s->d.prev = &s->d;
	s->d.next = &s->d;
	s->nd = 0;
	pool_init(&s->dpool);
	
	roll_init(&s->pixroll, c);
	
	s->logtime = s->curtime;
	s->recv_frames = 0;
	s->sent_frames = 0;
	s->sent_bytes = 0;
	s->sent_bytes_0 = 1;
	
	s->switch_time = s->curtime;
	s->switch_frame = s->recv_frames;
	s->wakeup_delay = c->vwakeup;
	
	s->yuv_buff_size = 0;
	s->bkg_buff_size = 0;
	s->jpeg_buff_size = 0;
	s->yuv_buff = NULL;
	s->bkg_buff = NULL;
	s->frg_buff = NULL;
	s->jpeg_buff = NULL;
	
	return(0);
}

// ***********************************************************************
// 
// ***********************************************************************
static void server_clear(server *s)
{
	client *c, *next;

	log_time(" Clean exit.\n");
	
	if (s->fdev.fd >= 0) {
		log_time("  video_shutdown():\n");
		video_shutdown(&s->fdev);
		s->fdev.fd = -1;
	}
	
	mask_clear(&s->fmask);
	
	if (s->yuv_buff)
		free(s->yuv_buff);
	if (s->bkg_buff)
		free(s->bkg_buff);
	if (s->frg_buff)
		free(s->frg_buff);
	if (s->jpeg_buff)
		free(s->jpeg_buff);

	roll_clear(&s->pixroll);
	
	if (s->fcur) {
		refbuf_ref(s->fcur, -1);
		s->fcur = NULL;
	}
	
	c = s->d.next;
	
	while (c != &s->d) {
		next = c->next;
		client_destroy(s, c);
		c = next;
	}
	
	pool_clear(&s->fpool);
	pool_clear(&s->dpool);
	
	close(s->d.fd);
}


// ***********************************************************************
// 
// ***********************************************************************
static void server_stats(server *s)
{
	double sent_total;
	double sent_bandwidth;
	
	if ((s->fdev.fd < 0)
	// ||  (s->sent_bytes == s->sent_bytes_0)
	||  (s->logtime > s->curtime - s->conf->sinterval))
		return;

	sent_total = (1.0 / (1 << 20)) * s->sent_bytes;
	sent_bandwidth =
		((8.0 / (1 << 20)) * (s->sent_bytes - s->sent_bytes_0))
		/ (s->curtime - s->logtime);
	
	if (sent_bandwidth < 0.0)
		sent_bandwidth = 0.0;
		
	log_time("  frames recv: %d, comp: %d, sent: %d"
		" (%.3f MB, %.3f Mbps)\n",
		s->recv_frames, s->fpool.seq, s->sent_frames,
		sent_total, sent_bandwidth);

	s->sent_bytes_0 = s->sent_bytes;
	s->logtime = s->curtime;
}

// ***********************************************************************
// 
// ***********************************************************************
static int server_iter(server *s)
{
	struct pollfd *fds;
	client *c, *next;
	ssize_t n;
	int nfds, t, pr;
	
	// Create fds
	nfds = 2 + s->nd;
	fds = (struct pollfd *)malloc((size_t)nfds * sizeof(struct pollfd));
	
	if (!fds) {
		log_time("malloc(struct pollfd): %s\n",
			strerror(errno));
		return(-1);
	}
	
	// Fill fds
	fds[0].fd = s->d.fd;
	fds[0].events = POLLIN;
	fds[0].revents = 0;

	fds[1].fd = s->fdev.fd;
	fds[1].events = POLLIN;
	fds[1].revents = 0;
	
	t = 2;
	c = s->d.next;
	
	while (c != &s->d) {
		fds[t].fd = c->fd;
		fds[t].events = POLLIN;
		
		if (((c->hsize) && (c->hpos < c->hsize))
		||  ((c->fbuf) && (c->fpos < c->fbuf->size)))
			fds[t].events |= POLLOUT;

		fds[t].revents = 0;
		
		c = c->next;
		t++;
	}
	
	ASSERT(t == nfds);
	
	// poll()
	pr = poll(fds, (nfds_t)nfds, (int)(1000.0 * s->conf->pinterval));
	
	if (pr < 0) {
		log_time("poll(): %s\n", strerror(errno));
		free(fds);
		return(-1);
	}
	
	s->curtime = time(NULL);
	
	// Accept
	if (fds[0].revents & POLLIN) {
		client_create(s);
	}
	
	if (fds[0].revents & (~POLLIN)) {
		log_time("poll(): listen socket shut down\n");
		free(fds);
		return(-1);
	}
	
	// Video
	if (fds[1].revents & POLLIN) {
		if (frame_handle(s)) {
			log_time("  Warning: frame_handle() failure\n"
				"  video_shutdown():\n");
			video_shutdown(&s->fdev);
			s->fdev.fd = -1;
			fds[1].revents = 0;
			//free(fds);
			//return(-1);
		}
	}
	
	if (fds[1].revents & (~POLLIN)) {
		log_time("  Warning: poll(): v4l2 device shut down (%d)\n"
			"  video_shutdown():\n",
			fds[1].revents);
		video_shutdown(&s->fdev);
		s->fdev.fd = -1;
		fds[1].revents = 0;
		//free(fds);
		//return(-1);
	}

	// I/O
	t = 2;
	c = s->d.next;
	
	while (t < nfds) {
		ASSERT(fds[t].fd == c->fd);
		
		next = c->next;
		
		c->flags &= ~(CLIENT_IN | CLIENT_OUT);
		
		if (fds[t].revents & POLLIN) {
			n = read(c->fd, c->buff + c->pos,
				REQ_SIZE - c->pos - 1);
			
			if (n <= 0) {
				log_time("%s:%s: read(): %s\n",
					c->addr.host, c->addr.service,
					(n < 0) ? strerror(errno)
						: "Disconnect");
				c->flags |= CLIENT_KILL;
			} else {
				c->pos += (size_t)n;
				c->buff[c->pos] = 0;
				c->flags |= CLIENT_IN;
			}
		}
		
		if (fds[t].revents & POLLOUT) {
			if (c->hpos < c->hsize) {
				n = write(c->fd, c->hbuf + c->hpos,
					c->hsize - c->hpos);
			} else {
				n = write(c->fd,
					(char *)c->fbuf->data + c->fpos,
					c->fbuf->size - c->fpos);
			}
			
			if (n <= 0) {
				log_time("%s:%s: write(): %s\n",
					c->addr.host, c->addr.service,
					(n < 0) ? strerror(errno)
						: "Disconnect");
				c->flags |= CLIENT_KILL;
			} else {
				if (c->hpos < c->hsize)
					c->hpos += (size_t)n;
				else
					c->fpos += (size_t)n;
				
				c->flags |= CLIENT_OUT;
			}
		}
		
		if (fds[t].revents & (~POLLIN) & (~POLLOUT)) {
			c->flags |= CLIENT_KILL;
		}
		
		client_handle(s, c);
		
		c = next;
		t++;
	}
	
	// free fds
	free(fds);
	
	// wakeup
	if ((s->fdev.fd < 0)
	&&  (s->wakeup_delay >= 1)
	&&  (s->curtime > s->switch_time + s->wakeup_delay)) {
		log_time("  Wakeup.\n");
		s->fatime = s->curtime;
	}
	
	// v4l2 init/shutdown
	if ((s->fdev.fd < 0) && (s->fatime == s->curtime)) {
		log_time("  video_init():\n");
	
		if (video_init(&s->fdev, s->conf->vwidth, s->conf->vheight,
		VIDEO_YUYV,
		s->conf->vdev)) {
			s->fdev.fd = -1;
			log_time("  Warning: video_init() failed.\n");
			//return(-1);
		} else if (mask_setup(&s->fmask,
		s->fdev.width, s->fdev.height)) {
			log_time("  Warning: mask_setup() failed.\n");
			log_time("  video_shutdown():\n");
			video_shutdown(&s->fdev);
			s->fdev.fd = -1;
			return(-1);
		} else {
			if (s->conf->ipixels) {
				s->iw = (int)(sqrt((double)s->conf->ipixels
					* s->fmask.mw / s->fmask.mh));
				s->ih = (int)(sqrt((double)s->conf->ipixels
					* s->fmask.mh / s->fmask.mw));
			} else if ((!s->conf->iwidth) && (!s->conf->iheight)) {
				s->iw = s->fmask.mw;
				s->ih = s->fmask.mh;
			} else if (!s->conf->iheight) {
				s->iw = s->conf->iwidth;
				s->ih = s->fmask.mh * s->iw / s->fmask.mw;
			} else if (!s->conf->iwidth) {
				s->ih = s->conf->iheight;
				s->iw = s->fmask.mw * s->ih / s->fmask.mh;
			} else {
				s->iw = s->conf->iwidth;
				s->ih = s->conf->iheight;
			}
			
			s->iw &= (~7);
			s->ih &= (~7);
			
			log_time("Raw image:    %4d x %4d\n",
				s->fdev.width, s->fdev.height);
			log_time("Masked image: %4d x %4d = %d (at %d, %d)\n",
				s->fmask.mw, s->fmask.mh,
				s->fmask.mw * s->fmask.mh,
				s->fmask.mx, s->fmask.my);
			log_time("Scaled image: %4d x %4d = %d (target %d)\n",
				s->iw, s->ih, s->iw * s->ih, s->conf->ipixels);
			
			s->bkg_buff_size = (size_t)s->iw * (size_t)s->ih;
			s->yuv_buff_size = s->bkg_buff_size * 3;
			s->jpeg_buff_size = s->yuv_buff_size;
			if (s->jpeg_buff_size < 8192)
				s->jpeg_buff_size = 8192;
			
			s->yuv_buff = malloc(s->yuv_buff_size);
			s->bkg_buff = malloc(s->bkg_buff_size);
			s->frg_buff = malloc(s->bkg_buff_size);
			s->jpeg_buff = malloc(s->jpeg_buff_size);
			
			if ((!s->jpeg_buff)
			||  (!s->yuv_buff)
			||  (!s->bkg_buff)
			||  (!s->frg_buff)) {
				log_time("Video init: "
					"malloc(%zd + 2*%zd + %zd): %s\n",
					s->yuv_buff_size, s->bkg_buff_size,
					s->jpeg_buff_size, strerror(errno));
				return(-1);
			}
			
			memset(s->bkg_buff, 0, s->bkg_buff_size);
			memset(s->frg_buff, 0, s->bkg_buff_size);
			s->switch_time = s->curtime;
			s->switch_frame = s->recv_frames;
		}
	}
	
	if ((s->conf->vsleep >= 1)
	&&  (s->fdev.fd >= 0)
	&&  (s->fatime < s->curtime - s->conf->vsleep)) {
		log_time("  video_shutdown():\n");
		video_shutdown(&s->fdev);
		s->fdev.fd = -1;
		free(s->yuv_buff);
		free(s->bkg_buff);
		free(s->frg_buff);
		free(s->jpeg_buff);
		s->yuv_buff = NULL;
		s->jpeg_buff = NULL;
		
		if (s->fcur)
			refbuf_ref(s->fcur, -1);

		s->fcur = NULL;
		s->fready = 0;

		s->switch_time = s->curtime;
		s->switch_frame = s->recv_frames;
	}
	
	server_stats(s);

	// Debug clean exit:
	//return(s->sent_frames > 25);
	
	return(0);
}

// ***********************************************************************
// 
// ***********************************************************************
int  server_loop(config *c)
{
	server s;
	int r;
	pid_t pid;
	
	if (signal(SIGPIPE, SIG_IGN)) {
		log_time("signal(SIGPIPE, SIG_IGN): %s\n", strerror(errno));
		return(-1);
	}
	
	if (server_init(&s, c))
		return(-1);
	
	if (c->detach) {
		if (c->logfile) {
			log_file = fopen(c->logfile, "w");
			
			if (!log_file) {
				log_time("%s: %s\n",
					c->logfile, strerror(errno));
				return(-1);
			}
			
			setbuf(log_file, NULL);
		}
		
		pid = fork();
		
		if (pid < 0) {
			log_time("fork(): %s\n", strerror(errno));
			return(-1);
		}
		
		if (pid > 0)
			return(0);
		
		close(0);
		close(1);
		if (c->logfile)
			close(2);
	}
	
	do {
		r = server_iter(&s);
	} while (r == 0);
	
	server_clear(&s);
	
	return(r < 0);
}


